import { ApiMsgEnum, EntityTypeEnum, IClientInput, InputTypeEnum, IState, toFixed } from '../Common'
import type Player from './Player'
import PlayerManager from './PlayerManager'
import RoomManager from './RoomManager'

export default class Room {
  id: number
  players: Set<Player> = new Set()
  lastSyncTime?: number
  timers: NodeJS.Timer[] = []

  private inputs: Array<IClientInput> = []

  constructor(rid: number) {
    this.id = rid
  }


  join(uid: number) {
    const player = PlayerManager.Instance.getPlayerById(uid)
    if (player) {
      player.rid = this.id
      this.players.add(player)
    }

  }

  leave(uid: number) {
    const player = PlayerManager.Instance.getPlayerById(uid)
    if (player) {
      player.rid = -1
      this.players.delete(player)
      if (!this.players.size) {
        this.timers.forEach(t => clearInterval(t))
        RoomManager.Instance.closeRoom(this.id)
      }
    }
  }

  sync() {
    for (const player of this.players) {
      player.connection.sendMsg(ApiMsgEnum.MsgRoom, {
        room: RoomManager.Instance.getRoomView(this)
      })
    }
  }

  start() {
    const state: IState = {
      players: [...this.players].map((player, index) => ({
        id: player.id,
        nickname: player.nickname,
        position: {
          x: -150 + index * 300,
          y: -150 + index * 300,
        },
        direction: {
          x: 1,
          y: 0
        },
        hp: 100,
        type: EntityTypeEnum.Actor2,
        weaponType: EntityTypeEnum.Weapon2,
        bulletType: EntityTypeEnum.Bullet2,
      })),
      bullets: [],
      nextBulletId: 1
    }

    for (const player of this.players) {
      player.connection.sendMsg(ApiMsgEnum.MsgGameStart, {
        state
      })
    }
    this.listenPlayer()
    let t1 = setInterval(() => {
      this.syncInput()
    }, 100)
    let t2 = setInterval(() => {
      this.timePast()
    }, 16)
    this.timers = [t1, t2]
  }

  listenPlayer() {
    for (const player of this.players) {
      player.connection.listenMsg(ApiMsgEnum.MsgClientSync, (input) => {
        this.inputs.push(input)
      })
    }
  }

  syncInput() {
    const inputs = this.inputs
    this.inputs = []


    for (const player of this.players) {
      player.connection.sendMsg(ApiMsgEnum.MsgServerSync, inputs)
    }
  }

  timePast() {
    let now = process.uptime();
    const dt = now - (this.lastSyncTime ?? now)
    this.inputs.push({
      type: InputTypeEnum.TimePast,
      dt: toFixed(dt)
    })
    this.lastSyncTime = now;
  }
}
